/*
 * Copyright (c) 2002, Stuart Walsh
 * All rights reserved.
 *
 * This file is licensed under the terms of the GPL license.  See the file
 * COPYING for more information.
 *
 * */
 
#include "ac-config.h"

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/socket.h>
#ifndef NOSSL
#include <openssl/ssl.h>
#include <openssl/crypto.h>
#include <openssl/err.h>
#endif
#include <net/if.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>

#include "tunnel.h"
#include "config.h"

#define VERSION "0.0.1"

#ifndef NOSSL
SSL_CTX *ctx = NULL;
struct ssl_st *ssl;
#endif
struct conf *config;

int init_ssl(void);
void cleanup_and_exit(int rcode)
{
    free_conf(config);
    _exit(rcode);
}

int init_ssl(void)
{
#ifndef NOSSL
    SSL_METHOD *method;
    
    OpenSSL_add_all_algorithms();
    SSL_load_error_strings();
    method = TLSv1_client_method();
    ctx = SSL_CTX_new(method);
    if(ctx == NULL)
    {
        ERR_print_errors_fp(stderr);
        return -1;
    }
#endif
    return 0;
}

int connect_socket(char *host, char *port)
{
    struct addrinfo hints, *res;
    int err, sd;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;
    
    if((err = getaddrinfo(host, port, &hints, &res)))
    {
        fprintf(stderr, "ERROR: Cannot connect to host: %s\n", 
                gai_strerror(err));
        cleanup_and_exit(-1);
    }

    if(!res)
        return -1;

    sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if(sd < 0)
    {
        perror("ERROR: Unable to create socket");
        return -1;
    }
    if(connect(sd, res->ai_addr, res->ai_addrlen) < 0)
    {
        perror("ERROR: Unable to connect");
        return -1;
    }
    if(!config->local)
    {
        char addr[NI_MAXHOST];
        struct sockaddr_in sock;
        socklen_t len = sizeof(sock);
        if(getsockname(sd, (struct sockaddr*)&sock, &len) < 0)
        {
            perror("ERROR: Unable to get local address");
            return -1;
        }
        getnameinfo((struct sockaddr*)&sock, len, addr, NI_MAXHOST,
                NULL, 0, NI_NUMERICHOST);
        config->local = strdup(addr);
    }
    
#ifndef NOSSL
    ssl = SSL_new(ctx);
    if(!ssl)
    {
        ERR_print_errors_fp(stderr);
        return -1;
    }
    SSL_set_fd(ssl, sd);
    if(SSL_connect(ssl) < 0)
    {
        ERR_print_errors_fp(stderr);
        return -1;
    }
    if((err = SSL_get_verify_result(ssl))!=X509_V_OK)
        fprintf(stderr, "WARNING: Unable to verify Server Certificate\n");
#endif

    return sd;
}


void send_updates(int fd)
{
    int ret;
    char buffer[256];

    memset(buffer, 0, 255);
    snprintf(buffer, 255, "USER %s\r\nPASS %s\r\nUPIP %s\r\nQUIT\r\n",
            config->user, config->pass, config->local);

#ifdef NOSSL
    ret = write(fd, buffer, strlen(buffer));
    if(ret < 0)
    {
        perror("ERROR: write()");
        return;
    }
    ret = read(fd, buffer, 255);
    if(ret < 0)
    {
        perror("ERROR: read()");
        return;
    }
    buffer[ret] = '\0';
    printf("%s\n", buffer);
#else
    ret = SSL_write(ssl, buffer, strlen(buffer));
    if(ret < 0)
    {
        ERR_print_errors_fp(stderr);
        return;
    }
    ret = SSL_read(ssl, buffer, 255);
    while(ret > 0)
    {
        buffer[ret] = '\0';
        printf("%s", buffer);
        ret = SSL_read(ssl, buffer, 255);
    }
    if(ret < 0)
    {
        ERR_print_errors_fp(stderr);
        return;
    }
#endif
}

int main(int argc, char **argv)
{
    int fd;
    char path[256];

    strncpy(path, CPATH, 255);
    strncat(path, "tunnelc.conf", 255);
    
    if((config = calloc(1, sizeof(struct conf))) == NULL)
    {
        perror("ERROR: Unable to allocate memory");
        exit(-1);
    }
    if(process_conf(config, path) == 0) 
        cleanup_and_exit(-1);
    if(init_ssl())
    {
        fprintf(stderr, "ERROR: Unable to initialise SSL.\n");
        cleanup_and_exit(-1);
    }
    fd = connect_socket(config->host, config->port);
    if(fd < 0)
        exit(-1);

#if 0    
    if(!config->local)
    {
        ip = get_ip_from_interface(config->linterface);
        if(!ip)
        {
            fprintf(stderr, "ERROR: Unable to get local ip\n");
            cleanup_and_exit(-1);
        }
        config->local = strdup(ip);
    }
#endif

    send_updates(fd);
    if(update_tunnel(config->tinterface, config->local, config->remote))
    {
        if(errno == ENXIO || errno == ENODEV)
        {
            char buffer[512];
            if(create_tunnel(config->tinterface, config->local, config->remote))
                cleanup_and_exit(-1);
            if(add_endpoints(config->tinterface, config->myend, config->hisend))
                cleanup_and_exit(-1);
#ifdef LINUX
            snprintf(buffer, 511, "route -A inet6 add 2000::/3 dev %s",
                    config->tinterface);
#else
            snprintf(buffer, 511, "route add -inet6 default %s", config->hisend);
#endif
            if(system(buffer) < 0)
            {
                perror("ERROR: Unable to add route\n");
                cleanup_and_exit(-1);
            }
            printf("Tunnel(%s) created\n", config->tinterface);
        }
        else
            perror("Unable to update interface");
    }
    else
        printf("Tunnel(%s) endpoint updated to %s\n", config->tinterface, 
                config->local);
    
    cleanup_and_exit(0);
    return 0;
}
